// SPDX-License-Identifier: MIT
/* simpler mkimage with minimum options */

#define _GNU_SOURCE
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <stdio.h>
#include <sys/stat.h>
#include <getopt.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdarg.h>
#include <errno.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <zlib.h>

/* global variable used for cleanup:
 * if we created the output path and fail somewhere in between we
 * should clean it up
 */
char *out_path;

/* from uboot include/image.h */
#define IH_NMLEN 32
/*
 * Legacy format image header,
 * all data in network byte order (aka natural aka bigendian).
 */
struct image_header {
	uint32_t ih_magic; /* Image Header Magic Number    */
	uint32_t ih_hcrc; /* Image Header CRC Checksum    */
	uint32_t ih_time; /* Image Creation Timestamp     */
	uint32_t ih_size; /* Image Data Size              */
	uint32_t ih_load; /* Data  Load  Address          */
	uint32_t ih_ep; /* Entry Point Address          */
	uint32_t ih_dcrc; /* Image Data CRC Checksum      */
	uint8_t ih_os; /* Operating System             */
	uint8_t ih_arch; /* CPU architecture             */
	uint8_t ih_type; /* Image Type                   */
	uint8_t ih_comp; /* Compression Type             */
	uint8_t ih_name[IH_NMLEN]; /* Image Name           */
} __attribute__((packed));

/* data itself is prefixed by an array of u32 sizes terminated by 0
 * we only have one file here
 */
#define DATA_HEADER_SIZE 8

static void cleanup_exit(void)
{
	if (out_path)
		unlink(out_path);
	exit(EXIT_FAILURE);
}

static void fatal_fprintf(FILE *out, const char *restrict fmt, ...)
{
	va_list ap;
	int ret, saved_errno;

	va_start(ap, fmt);
	ret = vfprintf(out, fmt, ap);
	saved_errno = errno;
	va_end(ap);

	if (ret < 0) {
		errno = saved_errno;
		perror("vfprintf");
		cleanup_exit();
	}
}

static void print_usage(FILE *out, char *progname)
{
	fatal_fprintf(out, "Usage: %s [-q] script.txt [script.scr]\n",
		      progname);
	fatal_fprintf(out,
		      "Equivalent to mkimage -T script -C none -a 0 -e 0 \\\n");
	fatal_fprintf(
		out,
		"                      -n \"boot script\" -d script.txt bootscript\n");
	fatal_fprintf(out, "\n");
	fatal_fprintf(out, "Options:\n");
	fatal_fprintf(out, "    -q: quiet output\n");
	fatal_fprintf(out, "\n");
	fatal_fprintf(
		out,
		"If output name is omitted, source path with .scr extension is used\n");
}

static size_t read_file(const char *path, char **data)
{
	int fd = open(path, O_RDONLY);
	if (fd < 0) {
		fatal_fprintf(stderr, "%s: %m\n", path);
		cleanup_exit();
	}

	struct stat sb;
	if (fstat(fd, &sb) < 0) {
		perror("fstat source text");
		cleanup_exit();
	}

	char *text = malloc(sb.st_size + DATA_HEADER_SIZE);
	if (!text) {
		perror("malloc source text");
		cleanup_exit();
	}

	/* real file size here, return topped up size */
	((uint32_t *)text)[0] = htobe32(sb.st_size);
	((uint32_t *)text)[1] = 0;

	/* assume no short reads */
	ssize_t n = read(fd, text + DATA_HEADER_SIZE, sb.st_size);
	if (n < 0) {
		perror("read source text");
		cleanup_exit();
	}
	if (n != sb.st_size) {
		fatal_fprintf(stderr, "short read for source text\n");
		cleanup_exit();
	}

	/* Use strlen to check for early nul byte -- it's not forbidden
	 * per se but there should be no valid use for these except getting
	 * arguments wrong
	 */
	size_t len = strlen(text + DATA_HEADER_SIZE);
	if (len != (size_t)sb.st_size) {
		fatal_fprintf(
			stderr,
			"source file %s contained a NUL byte at %zd (of %ld), wrong argument?\n",
			path, len, sb.st_size);
		cleanup_exit();
	}

	*data = text;
	return sb.st_size + DATA_HEADER_SIZE;
}

static void write_full(int fd, void *data, ssize_t size)
{
	ssize_t n = write(fd, data, size);
	if (n < 0) {
		perror("write data");
		cleanup_exit();
	}
	if (n != size) {
		fatal_fprintf(stderr, "short write for script data\n");
		cleanup_exit();
	}
}

int main(int argc, char *argv[])
{
	uint32_t fake_timestamp = 0;
	bool quiet = false;

	while (true) {
		static const struct option long_options[] = {
			{ "help", no_argument, NULL, 'h' },
			{ "quiet", no_argument, NULL, 'q' },
			{ "timestamp", required_argument, NULL, 'T' },
			{ NULL, 0, NULL, 0 }
		};

		int c = getopt_long(argc, argv, "hT:q", long_options, NULL);
		if (c == -1) {
			break;
		}

		switch (c) {
		case 'h':
			print_usage(stdout, argv[0]);
			exit(EXIT_SUCCESS);
		case 'q':
			quiet = true;
			break;
		case 'T':
			fake_timestamp = strtol(optarg, NULL, 0);
			break;
		case '?':
			print_usage(stderr, argv[0]);
			exit(EXIT_FAILURE);
		default:
			fatal_fprintf(
				stderr,
				"?? getopt returned character code 0%o ??\n",
				c);
			exit(EXIT_FAILURE);
		}
	}

	if (argc - optind > 2 || argc - optind < 1) {
		print_usage(stderr, argv[0]);
		exit(EXIT_FAILURE);
	}
	/* we don't care about big images, just read full script in memory */
	char *text_path = argv[optind], *text;
	size_t text_size = read_file(text_path, &text);

	if (argc - optind == 2) {
		out_path = strdup(argv[optind + 1]);
	} else {
		char *dot = strrchr(text_path, '.');
		char *slash = strrchr(text_path, '/');
		if (!dot || (slash && dot < slash)) {
			dot = text_path + strlen(text_path);
		}
		size_t len = dot - text_path;
		out_path = malloc(len + 4 /* scr\0 */);
		if (!out_path) {
			perror("malloc");
			exit(EXIT_FAILURE);
		}
		strncpy(out_path, text_path, len);
		strcpy(out_path + len, ".scr");
	}
	int script_fd = open(out_path, O_WRONLY | O_TRUNC | O_CREAT, 0666);
	if (script_fd < 0) {
		perror("open destination script");
		exit(EXIT_FAILURE);
	}

	/* actual mkimage */
	struct image_header header = { 0 };
	header.ih_magic = 0x56190527;
	header.ih_time = htobe32(fake_timestamp ?: time(NULL));
	header.ih_size = htobe32(text_size);
	header.ih_dcrc = htobe32(crc32(0, (uint8_t *)text, text_size));
	header.ih_type = 6; /* IH_TYPE_SCRIPT */
	strcpy((char *)header.ih_name, "boot script");

	/* must be last with crc previously zeroed out */
	header.ih_hcrc = htobe32(crc32(0, (uint8_t *)&header, sizeof(header)));
	write_full(script_fd, &header, sizeof(header));
	write_full(script_fd, text, text_size);

	if (!quiet)
		fatal_fprintf(stdout, "Wrote %s\n", out_path);

	free(out_path);
	free(text);
	return EXIT_SUCCESS;
}
